﻿using System;
using System.Collections.Generic;
using System.Text;

namespace BeeDevelopment.Infrared {

	/// <summary>
	/// Represents a stream of bits and provides methods for manipulation.
	/// </summary>
	public class BitStream : List<bool>, ICloneable, IEquatable<BitStream> {

		#region Constructors

		/// <summary>
		/// Creates an empty instance of a <see cref="BitStream"/>.
		/// </summary>
		public BitStream()
			: base() {
		}
		/// <summary>
		/// Creates an empty instance of a <see cref="BitStream"/>.
		/// </summary>
		/// <param name="capacity">The number of bits that the collection can initially store.</param>
		public BitStream(int capacity)
			: base(capacity) {
		}
		/// <summary>
		/// Creates an instance of a <see cref="BitStream"/> from an existing collection of bits.
		/// </summary>
		/// <param name="collection">The collection of bits to copy to the new <see cref="BitStream"/>.</param>
		public BitStream(IEnumerable<bool> collection)
			: base(collection) {
		}

		#endregion

		#region Methods

		/// <summary>
		/// Adds a bit to the least significant part of the <see cref="BitStream"/>.
		/// </summary>
		/// <param name="bit">The bit to add.</param>
		public void AddLeastSignificantBit(bool bit) {
			this.Insert(0, bit);
		}

		/// <summary>
		/// Adds a bit to the most significant part of the <see cref="BitStream"/>.
		/// </summary>
		/// <param name="bit">The bit to add.</param>
		public void AddMostSignificantBit(bool bit) {
			this.Add(bit);
		}

		/// <summary>
		/// Creates a copy of a range of bits from the <see cref="BitStream"/>.
		/// </summary>
		/// <param name="index">The zero-based index from the least-significant bit at which the range starts.</param>
		/// <param name="count">The number of bits in the range.</param>
		/// <returns>A range of bits starting from <paramref name="index"/>.</returns>
		public new BitStream GetRange(int index, int count) {
			return new BitStream(base.GetRange(index, count));
		}

		#endregion

		#region Utility

		/// <summary>Loads the <see cref="BitStream"/> instance from a numeric.</summary>
		/// <param name="value">The numeric value to load the instance from.</param>
		/// <param name="bits">The number of bits (up to 64) to copy from the numeric.</param>
		private void FromNumeric(long value, int bits) {
			this.Capacity = bits;
			this.Clear();
			for (int i = 0; i < bits; ++i) {
				this.Add((value & 1) != 0);
				value >>= 1;
			}
		}

		/// <summary>Converts the <see cref="BitStream"/> to a numeric value.</summary>
		/// <returns>A 64-bit integer containing the bits.</returns>
		private long ToNumeric() {
			long Result = 0;
			for (int i = this.Count - 1; i >= 0; --i) {
				Result <<= 1;
				if (this[i]) Result |= 1;
			}
			return Result;
		}

		/// <summary>Converts the <see cref="BitStream"/> into a binary string representation.</summary>
		/// <returns>A binary string representation of the <see cref="BitStream"/>.</returns>
		public override string ToString() {
			var Result = new StringBuilder(this.Count);
			for (int i = this.Count - 1; i >= 0; --i) {
				Result.Append(this[i] ? '1' : '0');
			}
			return Result.ToString();
		}

		/// <summary>Serves as a hash function for the <see cref="BitStream"/>.</summary>
		/// <returns>A hash code for this instance.</returns>
		public override int GetHashCode() {
			return base.GetHashCode();
		}

		#endregion

		#region Useful operators

		/// <summary>
		/// Complements the bits in a <see cref="BitStream"/>.
		/// </summary>
		/// <param name="stream">The <see cref="BitStream"/> to complement.</param>
		/// <returns>A complemented version of <paramref="stream"/>.</returns>
		public static BitStream operator ~(BitStream stream) {
			return new BitStream(Array.ConvertAll(stream.ToArray(), n => !n));
		}

		#endregion

		#region Equality operators

		/// <summary>Compares this <see cref="BitStream"/> instance with an <see cref="Object"/> for equality.</summary>
		/// <param name="obj">The object to compare to this one.</param>
		/// <returns>True if <paramref="obj"/> is equal to this <see cref="BitStream"/>, false otherwise.</returns>
		public override bool Equals(object obj) {
			return obj is BitStream && (BitStream)obj == this;
		}

		/// <summary>Compares this <see cref="BitStream"/> instance with another for equality.</summary>
		/// <param name="obj">The <see cref="BitStream"/> to compare to this one.</param>
		/// <returns>True if <paramref="other"/> is equal to this <see cref="BitStream"/>, false otherwise.</returns>
		public bool Equals(BitStream other) {
			return other == this;
		}

		/// <summary>Compares two <see cref="BitStream"/> instances for equality.</summary>
		/// <param name="a">An <see cref="BitStream"/> instance to check.</param>
		/// <param name="b">An <see cref="BitStream"/> instance to check.</param>
		/// <returns>True if <paramref name="a"/> is equal to <paramref name="b"/>, false otherwise.</returns>
		public static bool operator ==(BitStream a, BitStream b) {
			if ((object)a == (object)b) return true;
			if ((object)a == null || (object)b == null || a.Count != b.Count) return false;
			for (int i = 0; i < a.Count; ++i) {
				if (a[i] != b[i]) return false;
			}
			return true;
		}

		/// <summary>Compares two <see cref="BitStream"/> instances for inequality.</summary>
		/// <param name="a">An <see cref="BitStream"/> instance to check.</param>
		/// <param name="b">An <see cref="BitStream"/> instance to check.</param>
		/// <returns>True if <paramref name="a"/> is not equal to <paramref name="b"/>, false otherwise.</returns>
		public static bool operator !=(BitStream a, BitStream b) {
			return !(a == b);
		}

		#endregion

		#region Implicit conversions from numeric types.

		/// <summary>Convert a <see cref="Int64"/> into a <see cref="BitStream"/>.</summary>
		/// <param name="value">The <see cref="Int64"/> to load into the <see cref="BitStream"/>.</param>
		/// <returns>A <see cref="BitStream"/> instance loaded from a <see cref="Int64"/>.</returns>
		public static implicit operator BitStream(long value) {
			var Result = new BitStream();
			Result.FromNumeric(value, 64);
			return Result;
		}
		/// <summary>Convert a <see cref="UInt64"/> into a <see cref="BitStream"/>.</summary>
		/// <param name="value">The <see cref="UInt64"/> to load into the <see cref="BitStream"/>.</param>
		/// <returns>A <see cref="BitStream"/> instance loaded from a <see cref="UInt64"/>.</returns>
		public static implicit operator BitStream(ulong value) {
			var Result = new BitStream();
			Result.FromNumeric((long)value, 64);
			return Result;
		}
		/// <summary>Convert a <see cref="Int32"/> into a <see cref="BitStream"/>.</summary>
		/// <param name="value">The <see cref="Int32"/> to load into the <see cref="BitStream"/>.</param>
		/// <returns>A <see cref="BitStream"/> instance loaded from a <see cref="Int32"/>.</returns>
		public static implicit operator BitStream(int value) {
			var Result = new BitStream();
			Result.FromNumeric(value, 32);
			return Result;
		}
		/// <summary>Convert a <see cref="UInt32"/> into a <see cref="BitStream"/>.</summary>
		/// <param name="value">The <see cref="UInt32"/> to load into the <see cref="BitStream"/>.</param>
		/// <returns>A <see cref="BitStream"/> instance loaded from a <see cref="UInt32"/>.</returns>
		public static implicit operator BitStream(uint value) {
			var Result = new BitStream();
			Result.FromNumeric(value, 32);
			return Result;
		}
		/// <summary>Convert a <see cref="Int16"/> into a <see cref="BitStream"/>.</summary>
		/// <param name="value">The <see cref="Int16"/> to load into the <see cref="BitStream"/>.</param>
		/// <returns>A <see cref="BitStream"/> instance loaded from a <see cref="Int16"/>.</returns>
		public static implicit operator BitStream(short value) {
			var Result = new BitStream();
			Result.FromNumeric(value, 16);
			return Result;
		}
		/// <summary>Convert a <see cref="UInt16"/> into a <see cref="BitStream"/>.</summary>
		/// <param name="value">The <see cref="UInt16"/> to load into the <see cref="BitStream"/>.</param>
		/// <returns>A <see cref="BitStream"/> instance loaded from a <see cref="UInt16"/>.</returns>
		public static implicit operator BitStream(ushort value) {
			var Result = new BitStream();
			Result.FromNumeric(value, 16);
			return Result;
		}
		/// <summary>Convert a <see cref="Byte"/> into a <see cref="BitStream"/>.</summary>
		/// <param name="value">The <see cref="Byte"/> to load into the <see cref="BitStream"/>.</param>
		/// <returns>A <see cref="BitStream"/> instance loaded from a <see cref="Byte"/>.</returns>
		public static implicit operator BitStream(byte value) {
			var Result = new BitStream();
			Result.FromNumeric(value, 8);
			return Result;
		}
		/// <summary>Convert a <see cref="SByte"/> into a <see cref="BitStream"/>.</summary>
		/// <param name="value">The <see cref="SByte"/> to load into the <see cref="BitStream"/>.</param>
		/// <returns>A <see cref="BitStream"/> instance loaded from a <see cref="SByte"/>.</returns>
		public static implicit operator BitStream(sbyte value) {
			var Result = new BitStream();
			Result.FromNumeric(value, 8);
			return Result;
		}

		#endregion

		#region Explicit conversions to numeric types.

		/// <summary>Convert a <see cref="BitStream"/> into a <see cref="Int64"/>.</summary>
		/// <param name="value">The <see cref="BitStream"/> to convert.</param>
		/// <returns>The converted <see cref="Int64"/>.</returns>
		public static explicit operator long(BitStream value) {
			return value.ToNumeric();
		}
		/// <summary>Convert a <see cref="BitStream"/> into a <see cref="UInt64"/>.</summary>
		/// <param name="value">The <see cref="BitStream"/> to convert.</param>
		/// <returns>The converted <see cref="UInt64"/>.</returns>
		public static explicit operator ulong(BitStream value) {
			return (ulong)value.ToNumeric();
		}
		/// <summary>Convert a <see cref="BitStream"/> into a <see cref="Int32"/>.</summary>
		/// <param name="value">The <see cref="BitStream"/> to convert.</param>
		/// <returns>The converted <see cref="Int32"/>.</returns>
		public static explicit operator int(BitStream value) {
			return (int)value.ToNumeric();
		}
		/// <summary>Convert a <see cref="BitStream"/> into a <see cref="UInt32"/>.</summary>
		/// <param name="value">The <see cref="BitStream"/> to convert.</param>
		/// <returns>The converted <see cref="UInt32"/>.</returns>
		public static explicit operator uint(BitStream value) {
			return (uint)value.ToNumeric();
		}
		/// <summary>Convert a <see cref="BitStream"/> into a <see cref="Int16"/>.</summary>
		/// <param name="value">The <see cref="BitStream"/> to convert.</param>
		/// <returns>The converted <see cref="Int16"/>.</returns>
		public static explicit operator short(BitStream value) {
			return (short)value.ToNumeric();
		}
		/// <summary>Convert a <see cref="BitStream"/> into a <see cref="UInt16"/>.</summary>
		/// <param name="value">The <see cref="BitStream"/> to convert.</param>
		/// <returns>The converted <see cref="UInt16"/>.</returns>
		public static explicit operator ushort(BitStream value) {
			return (ushort)value.ToNumeric();
		}
		/// <summary>Convert a <see cref="BitStream"/> into a <see cref="Byte"/>.</summary>
		/// <param name="value">The <see cref="BitStream"/> to convert.</param>
		/// <returns>The converted <see cref="Byte"/>.</returns>
		public static explicit operator byte(BitStream value) {
			return (byte)value.ToNumeric();
		}
		/// <summary>Convert a <see cref="BitStream"/> into a <see cref="SByte"/>.</summary>
		/// <param name="value">The <see cref="BitStream"/> to convert.</param>
		/// <returns>The converted <see cref="SByte"/>.</returns>
		public static explicit operator sbyte(BitStream value) {
			return (sbyte)value.ToNumeric();
		}

		#endregion

		#region ICloneable Members

		/// <summary>Creates a copy of the <see cref="BitStream"/>.</summary>
		/// <returns>A copy of the <see cref="BitStream"/>.</returns>
		public object Clone() {
			return new BitStream(this);
		}

		#endregion

	}
}