﻿using System;
using System.Diagnostics;
using System.Collections.Generic;

namespace BeeDevelopment.Infrared.Protocols {

	/// <summary>
	/// Decodes incoming NEC commands.
	/// </summary>
	class NecProtocolHandler : IRemoteControlProtocolHandler {

		#region Constants

		/// <summary>
		/// The minimum start bit length (nominally 9mS).
		/// </summary>
		const double MinimumStartBurstLength = 8.5E-3d;
		
		/// <summary>
		/// The maximum start bit length (nominally 9mS).
		/// </summary>
		const double MaximumStartBurstLength = 9.5E-3d;

		/// <summary>
		/// The minimum repeat space length (nominally 2.25mS).
		/// </summary>
		const double MinimumRepeatSpaceLength = 1.75E-3d;

		/// <summary>
		/// The maximum repeat space length (nominally 2.25mS).
		/// </summary>
		const double MaximumRepeatSpaceLength = 2.75E-3d;

		/// <summary>
		/// The threshold length between a short bit and a long bit.
		/// </summary>
		const double DataBitLengthThreshold = (1.125E-3d + 2.25E-3d) / 2.0d;

		/// <summary>
		/// The maximum time length between repeating commands (nominally 110mS - repeat pulse).
		/// </summary>
		const double RepeatCommandMaxLength = 130E-3;

		#endregion

		#region Private Fields

		private double BurstLength = 0;

		private bool ReceivingStartBit = false;

		/// <summary>
		/// Stores the incoming decoded bits.
		/// </summary>
		private NecCommand Command = null;

		/// <summary>
		/// Stores the last received command.
		/// </summary>
		private NecCommand LastCommand = null;

		/// <summary>
		/// Stores the number of times the received command has been repeated.
		/// </summary>
		private int LastCommandRepeatCount = 0;

		/// <summary>
		/// A <see cref="Stopwatch"/> instance used to time repeating commands.
		/// </summary>
		private Stopwatch RepeatTimer = null;

		#endregion

		#region Constructor

		/// <summary>
		/// Create an instance of the <see cref="NecProtocolHandler"/>.
		/// </summary>
		public NecProtocolHandler() {
			this.RepeatTimer = new Stopwatch();
			this.RepeatTimer.Start();
		}

		#endregion

		#region Protocol handler

		/// <summary>
		/// Handle the change in input pin signal state.
		/// </summary>
		/// <param name="time">The time, in seconds, since the pin last changed state.</param>
		/// <param name="newState">The new state of the pin.</param>
		public void PinChanged(double time, bool newState) {
			if (newState) {
				// We've just received a burst.
				this.BurstLength = time;
				if (time > NecProtocolHandler.MinimumStartBurstLength && time < NecProtocolHandler.MaximumStartBurstLength) {
					this.Command = new NecCommand();
					this.ReceivingStartBit = true;
				} else {
					this.ReceivingStartBit = false;
				}
			} else {
				// We've just received a space.
				if (this.ReceivingStartBit) {
					// We just received a start burst (9ms). Is it a repeater?
					if (this.LastCommand != null && time > NecProtocolHandler.MinimumRepeatSpaceLength && time < NecProtocolHandler.MaximumRepeatSpaceLength) {
						// If the repeat signal hasn't come in within a sensible time, scrap the last command.
						if ((double)this.RepeatTimer.ElapsedTicks / (double)Stopwatch.Frequency > NecProtocolHandler.RepeatCommandMaxLength) {
							this.LastCommand = null;
						} else {
							// It's a repeater!
							this.OnCommandReceived(new InfraredCommandReceivedEventArgs(this.LastCommand, ++this.LastCommandRepeatCount));
							this.RepeatTimer.Reset();
							this.RepeatTimer.Start();
						}
					}
				} else {
					// This is a regular data bit.
					double PulseLength = this.BurstLength + time;
					if (this.Command != null) {
						// Add the data bit.
						this.Command.AddMostSignificantBit(PulseLength > NecProtocolHandler.DataBitLengthThreshold);
						// Have we received everything?
						if (this.Command.Count == 32) {
							// Is everything shipshape?
							if (this.Command.IsValid) {
								this.LastCommand = this.Command;
								this.OnCommandReceived(new InfraredCommandReceivedEventArgs(this.Command, this.LastCommandRepeatCount = 1));
								this.RepeatTimer.Reset();
								this.RepeatTimer.Start();
							}
							this.Command = null;
						}
					}
				}
			}
		}

		#endregion

		#region Events
		
		public event EventHandler<InfraredCommandReceivedEventArgs> CommandReceived;

		protected virtual void OnCommandReceived(InfraredCommandReceivedEventArgs e) {
			if (this.CommandReceived != null) this.CommandReceived(this, e);
		}
		
		#endregion

	}
}