﻿using System;
using System.Diagnostics;
using System.Collections.Generic;

namespace BeeDevelopment.Infrared.Protocols {
	/// <summary>
	/// Decodes incoming new Panasonic commands.
	/// </summary>
	/// <remarks>See http://www.hifi-remote.com/forums/viewtopic.php?p=73415 for more information.</remarks>
	class NewPanasonicProtocolHandler : IRemoteControlProtocolHandler {

		#region Constants

		/// <summary>
		/// The minimum start bit length (3500uS).
		/// </summary>
		private const double StartBitMinLength = 3300E-6;

		/// <summary>
		/// The maximum start bit length (3500uS).
		/// </summary>
		private const double StartBitMaxLength = 3700E-6;

		/// <summary>
		/// The maximum data bit length (1310uS).
		/// </summary>
		private const double DataBitMaxLength = 1900E-6;

		/// <summary>
		/// The data bit length threshold between a 0 and a 1.
		/// </summary>
		private const double DataBitThresholdLength = 875E-6;

		/// <summary>
		/// The maximum time length between repeating commands.
		/// </summary>
		private const double RepeatCommandMaxLength = 200.0E-3;

		#endregion

		#region Private Fields

		/// <summary>
		/// Set to <c>true</c> when receiving a start bit, <c>false</c> otherwise.
		/// </summary>
		private bool ReceivingStartBit = false;

		/// <summary>
		/// Stores the command as it gets built up.
		/// </summary>
		private NewPanasonicCommand Command = null;

		/// <summary>
		/// Stores the last received command.
		/// </summary>
		private NewPanasonicCommand LastCommand = null;

		/// <summary>
		/// Stores the number of times the received command has been repeated.
		/// </summary>
		private int LastCommandRepeatCount = 0;

		/// <summary>
		/// A <see cref="Stopwatch"/> instance used to time repeating commands.
		/// </summary>
		private Stopwatch RepeatTimer = null;

		#endregion

		public NewPanasonicProtocolHandler() {
			this.RepeatTimer = new Stopwatch();
			this.RepeatTimer.Start();
		}

		public void PinChanged(double time, bool newState) {
			if (newState) {
				// We've just received a burst.
				if (this.ReceivingStartBit && time > NewPanasonicProtocolHandler.StartBitMinLength && time < NewPanasonicProtocolHandler.StartBitMaxLength) {
					// We've just received a start bit.
					this.Command = new NewPanasonicCommand();
				}
			} else {
				// We've just received a space.
				if (this.Command != null && time < NewPanasonicProtocolHandler.DataBitMaxLength) {
					this.Command.AddMostSignificantBit(time > NewPanasonicProtocolHandler.DataBitThresholdLength);
					if (this.Command.Count == 49) {
						var ReceivedCommand = this.Command;
						ReceivedCommand.RemoveAt(0);
						this.Command = null;
						if (ReceivedCommand.IsValid) {
							if (ReceivedCommand != this.LastCommand || (double)this.RepeatTimer.ElapsedTicks / (double)Stopwatch.Frequency > NewPanasonicProtocolHandler.RepeatCommandMaxLength) {
								this.LastCommand = ReceivedCommand;
								this.LastCommandRepeatCount = 0;
							}
							this.OnCommandReceived(new InfraredCommandReceivedEventArgs(ReceivedCommand, ++this.LastCommandRepeatCount));
							this.RepeatTimer.Reset();
							this.RepeatTimer.Start();
						}
					}
				} else {
					this.Command = null;
					this.ReceivingStartBit = (time > NewPanasonicProtocolHandler.DataBitMaxLength);
				}
			}
		}

		public event EventHandler<InfraredCommandReceivedEventArgs> CommandReceived;

		protected virtual void OnCommandReceived(InfraredCommandReceivedEventArgs e) {
			if (this.CommandReceived != null) this.CommandReceived(this, e);
		}

	}
}
