﻿using System;
using System.Diagnostics;
using System.Collections.Generic;

namespace BeeDevelopment.Infrared.Protocols {

	/// <summary>
	/// Decodes incoming RC5 commands.
	/// </summary>
	class RC5ProtocolHandler : IRemoteControlProtocolHandler {

		#region Constants

		/// <summary>
		/// The minimum pulse length within a bit (nominally 889uS).
		/// </summary>
		const double MinimumBitPulseLength = 800E-6d;
		
		/// <summary>
		/// The maximum pulse length within a bit (nominally 889uS*2).
		/// </summary>
		const double MaximumBitPulseLength = 980E-6d * 2.0d;

		/// <summary>
		/// The threshold between a "short" and "long" pulse.
		/// </summary>
		const double BitPulseLengthThreshold = (MinimumBitPulseLength + MaximumBitPulseLength) / 2.0d;

		/// <summary>
		/// The maximum time length between repeating commands.
		/// </summary>
		const double RepeatCommandMaxLength = 250E-3;

		#endregion

		#region Private Fields

		/// <summary>
		/// Stores the incoming decoded bits.
		/// </summary>
		private RC5Command Command = null;

		/// <summary>
		/// Set to true if we're waiting for a half-bit pulse to complete a decoded data bit.
		/// </summary>
		private bool WaitingHalfBit = false;

		/// <summary>
		/// Stores the last received command.
		/// </summary>
		private RC5Command LastCommand = null;

		/// <summary>
		/// Stores the number of times the received command has been repeated.
		/// </summary>
		private int LastCommandRepeatCount = 0;

		/// <summary>
		/// A <see cref="Stopwatch"/> instance used to time repeating commands.
		/// </summary>
		private Stopwatch RepeatTimer = null;

		#endregion

		#region Constructor

		/// <summary>
		/// Create an instance of the <see cref="RC5ProtocolHandler"/>.
		/// </summary>
		public RC5ProtocolHandler() {
			this.RepeatTimer = new Stopwatch();
			this.RepeatTimer.Start();
		}

		#endregion

		#region Protocol handler

		/// <summary>
		/// Handle the change in input pin signal state.
		/// </summary>
		/// <param name="time">The time, in seconds, since the pin last changed state.</param>
		/// <param name="newState">The new state of the pin.</param>
		public void PinChanged(double time, bool newState) {
			// Is the time out of bounds of a half-bit pulse? If so, cancel incoming data.
			if (time < RC5ProtocolHandler.MinimumBitPulseLength || time > RC5ProtocolHandler.MaximumBitPulseLength) {
				this.Command = new RC5Command(); // We start from one to include the start bit.
				this.Command.AddLeastSignificantBit(true);
				this.WaitingHalfBit = false;
			} else if (this.Command!=null && this.Command.Count < 14) {
				if (time > RC5ProtocolHandler.BitPulseLengthThreshold) {
					// We've stepped through into a whole bit.
					this.WaitingHalfBit = false;
					this.Command.AddLeastSignificantBit(!newState);
				} else {
					// We've only received half a bit.
					if (this.WaitingHalfBit) {
						this.Command.AddLeastSignificantBit(!newState);
						this.WaitingHalfBit = false;
					} else {
						this.WaitingHalfBit = true;
					}
				}

				// Have we received all 14 bits?
				if (this.Command.Count == 14) {

					// Process the received command.
					var ReceivedCommand = this.Command;
					this.Command = null;

					// Calculate the repeat time elapsed.
					double RepeatTimeSeconds = (double)this.RepeatTimer.ElapsedTicks / (double)Stopwatch.Frequency;

					// Is the command repeating?
					if (this.LastCommand != ReceivedCommand || RepeatTimeSeconds > RC5ProtocolHandler.RepeatCommandMaxLength) {
						this.LastCommandRepeatCount = 0;
						this.LastCommand = ReceivedCommand;
					}
					this.RepeatTimer.Reset();
					this.RepeatTimer.Start();

					// Raise the event.
					this.OnCommandReceived(new InfraredCommandReceivedEventArgs(ReceivedCommand, ++this.LastCommandRepeatCount));
				}
			}
		}

		#endregion

		#region Events
		
		public event EventHandler<InfraredCommandReceivedEventArgs> CommandReceived;

		protected virtual void OnCommandReceived(InfraredCommandReceivedEventArgs e) {
			if (this.CommandReceived != null) this.CommandReceived(this, e);
		}
		
		#endregion

	}
}