﻿using System;
using System.Diagnostics;

namespace BeeDevelopment.Infrared.Protocols {

	/// <summary>
	/// Decodes incoming SIRCS commands.
	/// </summary>
	class SircsProtocolHandler : IRemoteControlProtocolHandler {

		#region Constants

		/// <summary>
		/// The minimum time length for a start bit (nominally 2.4ms).
		/// </summary>
		private const double StartBitMinLength = 2.0E-3;

		/// <summary>
		/// Threshold time length between a "low" (0.6ms) and a "high" (1.2ms) bit.
		/// </summary>
		private const double DataBitLengthThreshold = 0.9E-3;

		/// <summary>
		/// The maximum time length between data bits. If this is exceeded, any data command transfer is cancelled.
		/// </summary>
		private const double IntraBitMaxLength = 0.8E-3;

		/// <summary>
		/// The maximum time length between repeating commands. Commands are supposed to repeat every 45ms.
		/// </summary>
		private const double RepeatCommandMaxLength = 120.0E-3;

		#endregion

		#region Private Fields

		/// <summary>
		/// Stores the command as it gets built up.
		/// </summary>
		private SircsCommand Command = null;

		/// <summary>
		/// Stores the last received command.
		/// </summary>
		private SircsCommand LastCommand = null;

		/// <summary>
		/// Stores the number of times the received command has been repeated.
		/// </summary>
		private int LastCommandRepeatCount = 0;

		/// <summary>
		/// A <see cref="Stopwatch"/> instance used to time repeating commands.
		/// </summary>
		private Stopwatch RepeatTimer = null;

		/// <summary>
		/// Allows the protocol handler to talk to its parent receiver.
		/// </summary>
		private RemoteControlReceiver Receiver;

		#endregion

		#region Constructor

		/// <summary>
		/// Create an instance of the <see cref="SircsProtocolHandler"/>.
		/// </summary>
		/// <param name="receiver">The <see cref="RemoteControlReceiver"/> feeding this instance.</param>
		public SircsProtocolHandler(RemoteControlReceiver receiver) {
			this.Receiver = receiver;
			this.RepeatTimer = new Stopwatch();
			this.RepeatTimer.Start();
		}

		#endregion

		#region Protocol handler

		/// <summary>
		/// Handle the change in input pin signal state.
		/// </summary>
		/// <param name="time">The time, in seconds, since the pin last changed state.</param>
		/// <param name="newState">The new state of the pin.</param>
		public void PinChanged(double time, bool newState) {
			if (newState) { // If the current signal level is high, we may assume that we've just timed a low pulse.

				// Have we received a start bit?
				if (time > SircsProtocolHandler.StartBitMinLength) {
					this.Command = new SircsCommand();
				} else if (this.Command != null) {
					// Process incoming bit.
					this.Command.AddMostSignificantBit(time > SircsProtocolHandler.DataBitLengthThreshold);
					// Have we received enough bits?
					switch (this.Command.Count) {
						case 12:
						case 15:
						case 20:
							// We've received enough bits to handle the input as a received command.
							// Check to see if there's any more data forthcoming.
							if (!this.Receiver.WaitForSignalLevel(false, SircsProtocolHandler.IntraBitMaxLength)) {

								var ReceivedCommand = this.Command;

								this.Command = null;

								// Calculate the repeat count.

								// Quickly grab the current time and current CTS level.
								long RepeatTimeTicks = this.RepeatTimer.ElapsedTicks;
								this.RepeatTimer.Reset();
								this.RepeatTimer.Start();

								// Calculate the repeat time elapsed.
								double RepeatTimeSeconds = (double)RepeatTimeTicks / (double)Stopwatch.Frequency;

								// Is the command repeating?
								if (ReceivedCommand == this.LastCommand && RepeatTimeSeconds < SircsProtocolHandler.RepeatCommandMaxLength) {
									++this.LastCommandRepeatCount;
								} else {
									this.LastCommandRepeatCount = 1;
									this.LastCommand = ReceivedCommand;
								}

								// Fire the event.
								this.OnCommandReceived(new InfraredCommandReceivedEventArgs(ReceivedCommand, this.LastCommandRepeatCount));
							}
							break;
					}
				}
			} else { // If the current signal level is low, we may assume that we've just timed a high pulse.
				// If a high pulse is too long, cancel any incoming commands.
				if (time > SircsProtocolHandler.IntraBitMaxLength) {
					this.Command = null;
				}
			}
		}

		#endregion

		#region Events
		
		public event EventHandler<InfraredCommandReceivedEventArgs> CommandReceived;

		protected virtual void OnCommandReceived(InfraredCommandReceivedEventArgs e) {
			if (this.CommandReceived != null) this.CommandReceived(this, e);
		}
		
		#endregion

	}
}