﻿using System;
using System.Diagnostics;
using System.IO.Ports;

namespace BeeDevelopment.Infrared {

	public class RemoteControlReceiver : IDisposable {

		#region Private Fields

		/// <summary>
		/// The serial port that the infrared receiver is connected to.
		/// </summary>
		private SerialPort Port;

		/// <summary>
		/// A timer used to keep track of when the input pin changed level and thus time pulses.
		/// </summary>
		private Stopwatch PinChangedTimer;

		// Protocol handlers
		private Protocols.SircsProtocolHandler SircsHandler;
		private Protocols.RC5ProtocolHandler RC5Handler;
		private Protocols.OldPanasonicProtocolHandler OldPanasonicHandler;
		private Protocols.NewPanasonicProtocolHandler NewPanasonicHandler;
		private Protocols.NecProtocolHandler NecHandler;
		private Protocols.JvcProtocolHandler JvcHandler;

		#endregion

		#region Construction / Destruction

		public RemoteControlReceiver(string portName) {
			// Set up the stopwatch.
			this.PinChangedTimer = new Stopwatch();
			this.PinChangedTimer.Start();

			// Set up the protocol handlers.
			this.SircsHandler = new Protocols.SircsProtocolHandler(this);
			this.SircsHandler.CommandReceived += (sender, e) => this.OnCommandReceived(e);
			this.RC5Handler = new Protocols.RC5ProtocolHandler();
			this.RC5Handler.CommandReceived += (sender, e) => this.OnCommandReceived(e);
			this.OldPanasonicHandler = new Protocols.OldPanasonicProtocolHandler();
			this.OldPanasonicHandler.CommandReceived += (sender, e) => this.OnCommandReceived(e);
			this.NewPanasonicHandler = new Protocols.NewPanasonicProtocolHandler();
			this.NewPanasonicHandler.CommandReceived += (sender, e) => this.OnCommandReceived(e);
			this.NecHandler = new Protocols.NecProtocolHandler();
			this.NecHandler.CommandReceived += (sender, e) => this.OnCommandReceived(e);
			this.JvcHandler = new Protocols.JvcProtocolHandler(this);
			this.JvcHandler.CommandReceived += (sender, e) => this.OnCommandReceived(e);

			// Set up the serial port.
			this.Port = new SerialPort(portName);
			this.Port.Open();
			this.Port.DtrEnable = this.Port.RtsEnable = true;
			this.Port.PinChanged += new SerialPinChangedEventHandler(PortPinChanged);
		}

		/// <summary>
		/// Release resources used by this class.
		/// </summary>
		public void Dispose() {
			if (this.Port != null) {
				this.Port.Dispose();
				this.Port = null;
			}
		}

		~RemoteControlReceiver() {
			this.Dispose();
		}

		#endregion

		#region Pin Handling

		/// <summary>
		/// Stores the last time the pin changed state in ticks.
		/// </summary>
		private long LastPinChangedTime = 0;

		/// <summary>
		/// Stores the old port pin state.
		/// </summary>
		private bool OldState;

		private void PortPinChanged(object sender, SerialPinChangedEventArgs e) {
			// Grab state and time quickly.
			bool NewState = this.Port.CtsHolding;
			long TimeTicks = this.PinChangedTimer.ElapsedTicks;
			// If it's the CTS pin that has changed, act on that change.
			if (e.EventType == SerialPinChange.CtsChanged) {
				double TimeSeconds = (double)(TimeTicks - this.LastPinChangedTime) / (double)Stopwatch.Frequency;
				this.LastPinChangedTime = TimeTicks;
				// Let all the protocol handlers know about the change.
				this.SircsHandler.PinChanged(TimeSeconds, NewState);
				this.RC5Handler.PinChanged(TimeSeconds, NewState);
				this.OldPanasonicHandler.PinChanged(TimeSeconds, NewState);
				this.NewPanasonicHandler.PinChanged(TimeSeconds, NewState);
				this.NecHandler.PinChanged(TimeSeconds, NewState);
				this.JvcHandler.PinChanged(TimeSeconds, NewState);
			}
			this.OldState = NewState;
		}

		/// <summary>
		/// Wait for the input pin to change to a particular level within a particular amount of time.
		/// </summary>
		/// <param name="level">The signal level the pin need to get to.</param>
		/// <param name="timeOut">The amount of time (in seconds) required to wait for.</param>
		/// <returns>True if the signal level changed to the one requested before timeout, false otherwise.</returns>
		internal bool WaitForSignalLevel(bool level, double timeOut) {
			long TimeOutTicks = this.PinChangedTimer.ElapsedTicks + (long)(timeOut * Stopwatch.Frequency);
			while (this.PinChangedTimer.ElapsedTicks < TimeOutTicks) {
				if (this.Port.CtsHolding == level) return true;
			}
			return false;
		}

		#endregion

		#region Events

		public event EventHandler<InfraredCommandReceivedEventArgs> CommandReceived;

		protected virtual void OnCommandReceived(InfraredCommandReceivedEventArgs e) {
			if (this.CommandReceived != null) this.CommandReceived(this, e);
		}

		#endregion

	}
}
